from clldutils.misc import deprecated

from cldfzenodo.api import API


def iter_records(community):
    deprecated('Use `API.iter_records` instead.')
    return API.iter_records(community=community, allversions=True)



from clldutils.misc import deprecated

from cldfzenodo.api import API


def iter_records(keyword, q=None, **kw):
    deprecated('Use `API.iter_records` instead.')
    return API.iter_records(keyword=keyword, allversions=True, _q=q)




def test_legacy(urlopen, mocker, recwarn):
    from cldfzenodo import (
        search_wordlists, search_structuredatasets, oai_cldf_datasets, oai_lexibank,
    )
    from cldfzenodo.oai import iter_records
    from cldfzenodo import search

    mocker.patch('cldfzenodo.api.urllib.request.urlopen', urlopen)

    res = list(search_wordlists(sort='x'))
    assert len(res) == 10
    assert recwarn.pop(UserWarning)
    assert recwarn.pop(DeprecationWarning)

    res = list(search_structuredatasets(sort='x'))
    assert len(res) == 10
    assert recwarn.pop(UserWarning)
    assert recwarn.pop(DeprecationWarning)

    res = list(oai_lexibank())
    assert len(res) == 10
    assert recwarn.pop(DeprecationWarning)

    res = list(oai_cldf_datasets())
    assert len(res) == 10
    assert recwarn.pop(DeprecationWarning)

    res = list(iter_records('x'))
    assert len(res) == 10
    assert recwarn.pop(DeprecationWarning)

    res = list(search.iter_records(None, q='x'))
    assert len(res) == 10
    assert recwarn.pop(DeprecationWarning)



import json
import logging

import pytest

from cldfzenodo.record import *


@pytest.fixture
def record_factory(fixtures_dir):
    def make_one(fname):
        return Record.from_dict(
            json.loads(fixtures_dir.joinpath(fname).read_text(encoding='utf8'))['hits']['hits'][0])
    return make_one


@pytest.fixture
def record(record_factory):
    return record_factory('record.json')


@pytest.mark.parametrize(
    'url_or_doi',
    [
        '10.5281/zenodo.5173799',
        "https://doi.org/10.5281/zenodo.5173799",
        "http://zenodo.org/doi/10.5281/zenodo.5173799",
        "https://zenodo.org/record/5173799",
        "https://zenodo.org/records/5173799",
    ]
)
def test_get_doi(url_or_doi):
    assert get_doi(url_or_doi) == '10.5281/zenodo.5173799'


def test_get_doi_error():
    with pytest.raises(ValueError):
        get_doi('http://example.com')


def test_GithubRepos():
    assert GithubRepos.from_url('http://example.com') is None
    repo = GithubRepos.from_url('https://github.com/org/repo')
    assert not repo.tag
    repo = GithubRepos.from_url('https://github.com/org/repo/tree/v1.0')
    assert repo.tag == 'v1.0'
    assert repo.clone_url
    assert repo.release_url


def test_Record(record):
    with pytest.raises(ValueError):
        _ = Record(doi='x', title='x')

    with pytest.raises(AssertionError):
        _ = Record(doi='10.5281/zenodo.4691101', title='')

    assert record.version == 'v1.1'
    assert record.concept_doi == '10.5281/zenodo.4691101'
    assert 'Greenhill, Simon J. and Haynie, Hannah J.' in record.bibtex

    rec = Record(doi='https://doi.org/10.5281/zenodo.4691101', title='', closed_access=True)
    assert rec.id == '4691101'


def test_Record_with_external_doi(record_factory):
    assert record_factory('record_with_external_doi.json')


def test_Record_multifile(record_factory, mocker, tmp_path):
    class Response:
        def __init__(self, *args):
            self.yielded = False
            self.code = 200

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def read(self):
            return b'abc'

    rec = record_factory('record_multifile.json')
    assert len(rec.download_urls) == 5
    mocker.patch('cldfzenodo.api.urllib.request.urlopen', Response)
    assert len(list(rec.download(tmp_path / 'test').iterdir())) == 5


def test_Record_download_dataset(tmp_path, mocker, fixtures_dir, caplog, record):
    class Response:
        def __init__(self, *args):
            self.yielded = False
            self.code = 200

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def read(self):
            return fixtures_dir.joinpath('petersonsouthasia-v1.1.zip').read_bytes()

    mocker.patch('cldfzenodo.api.urllib.request.urlopen', Response)

    with caplog.at_level(logging.INFO):
        assert record.download_dataset(tmp_path, log=logging.getLogger(__name__)).validate()
        assert caplog.records


def test_Record_citation(record, mocker):
    from cldfzenodo import API

    def urlopen(url):
        return mocker.Mock(
            read=lambda: "& The Citation".encode('utf8'))

    mocker.patch('cldfzenodo.api.urllib.request.urlopen', urlopen)
    assert record.get_citation(API) == '& The Citation'



import re
import json
import shlex
import typing
import subprocess
import urllib.error
import urllib.parse
import urllib.request

from pycldf import Source
from pycldf.ext.discovery import DatasetResolver
from clldutils.path import ensure_cmd

from cldfzenodo.record import Record, ZENODO_DOI_FORMAT, ZENODO_DOI_PATTERN, get_doi

__all__ = ['API', 'Api', 'ZenodoResolver']


def get_hits(res: typing.Union[str, dict, list]) -> typing.List[dict]:
    if isinstance(res, str):
        res = json.loads(res)
    return res if isinstance(res, list) else res['hits']['hits']


def q(**kw) -> str:
    """
    Put together a Zenodo search query from the data in `kw`.

    See https://help.zenodo.org/guides/search/ for details.
    """
    res = kw.pop('_q', '') or ''
    if kw:
        res += '' + ' '.join('{}:"{}"'.format(k, v) for k, v in kw.items())
    return res.strip()


class Results:
    def __init__(self,
                 api: 'Api',
                 allversions: typing.Optional[bool] = False,
                 community: typing.Optional[str] = None,
                 **params: str):
        """
        :param params: URL parameters (see https://developers.zenodo.org/#records)
        """
        self.api = api
        self.params = params
        self.community = community
        if allversions:
            self.params['allversions'] = 'true'
        self._page = 0
        self.more = True
        self.error = None

    def next(self) -> typing.List[Record]:
        self._page += 1
        params = {
            'sort': '-mostrecent',
            'page': str(self._page),
            'size': str(self.api.page_size)}
        params.update(self.params)
        try:
            hits = get_hits(self.api.records(community=self.community, params=params))
        except urllib.error.HTTPError as e:  # pragma: no cover
            # We want to be able to resume retrieval of results.
            hits = []
            self.error = e
            self._page -= 1
            raise
        self.more = len(hits) >= self.api.page_size
        return [Record.from_dict(r) for r in hits]


class Api:
    __base_url__ = "https://zenodo.org/api/"
    __communities__ = {}  # We cash community identifiers.

    def __init__(self, page_size=100):
        self.page_size = page_size

    def __call__(self,
                 what: str,
                 id_: typing.Optional[str] = None,
                 params: typing.Optional[typing.Dict[str, str]] = None,
                 headers: typing.Optional[typing.Dict[str, str]] = None,
                 _verbose: bool = False) -> str:
        """
        Perform an API call and return the response body as str.

        :param what: Name of the API resource, e.g. "records".
        :param id_: Optional identifier of a resource instance.
        :param params: URL query parameters.
        :param headers: HTTP request headers.
        :param _verbose:
        :return: Body of the response as text.
        """
        params = params or {}
        headers = headers or {}
        comps = urllib.parse.urlparse(
            self.__base_url__ + '{}{}'.format(what, '/' + id_ if id_ else ''))
        url = urllib.parse.urlunparse(list(comps[:3]) + ['', urllib.parse.urlencode(params), ''])
        req = urllib.request.Request(url)
        for k, v in headers.items():
            req.add_header(k, v)
        if _verbose:  # pragma: no cover
            print(url)
        return urllib.request.urlopen(req).read().decode('utf8')

    def records(self,
                id_: typing.Optional[str] = None,
                community: typing.Optional[str] = None,
                params: typing.Optional[typing.Dict[str, str]] = None,
                headers: typing.Optional[typing.Dict[str, str]] = None) -> str:
        """
        Perfom an API call to the records resource (possibly limited to one community).

        :param id_: Optional record identifier.
        :param community: Zenodo community identifier. If specified, only records in this community\
        are considered.
        :param params: URL query parameters.
        :param headers: HTTP request headers.
        :return: Body of the response as text.
        """
        what = ''
        if community:
            if community not in self.__communities__:
                res = get_hits(self('communities', params=dict(q=community)))
                self.__communities__[community] = res[0]['id']
            what = 'communities/{}/'.format(self.__communities__[community])
        what += 'records'
        return self(what, id_=id_, params=params, headers=headers)

    def iter_versions(self, conceptdoi: str) -> typing.Generator[Record, None, None]:
        """
        Retrieve all versions recorded for a concept DOI.
        """
        yield from self.iter_records(conceptdoi=get_doi(conceptdoi), allversions=True)

    @staticmethod
    def _first_record(res):
        hits = get_hits(res)
        if hits:
            return Record.from_dict(hits[0])

    def get_record(self,
                   doi: typing.Optional[str] = None,
                   conceptdoi: typing.Optional[str] = None,
                   version: typing.Optional[str] = None) -> typing.Union[Record, None]:
        """
        Retrieve a record specified by Zenodo DOI, given as URL or plain.

        :param doi: A Zenodo DOI or URL identifying a record (i.e. no concept DOI).
        :param conceptdoi: A Zenodo concept DOI. If no `version` is passed, the latest matching \
        record will be returned.
        :param version: A version tag, used to choose a particular record when only a concept doi \
        is passed.
        :return: A `Record` instance, if one is found, else `None`.
        """
        if doi:
            doi = get_doi(doi)
            assert not conceptdoi
            return self._first_record(self.records(params=dict(allversions='true', q=q(doi=doi))))
        assert conceptdoi
        conceptdoi = get_doi(conceptdoi)
        if not version:
            return self._first_record(self.records(params=dict(q=q(conceptdoi=conceptdoi))))
        for rec in self.iter_records(conceptdoi=conceptdoi, allversions=True):
            if rec.version.lstrip('v') == version.lstrip('v'):
                return rec

    def iter_records(self,
                     community: typing.Optional[str] = None,
                     keyword: typing.Optional[str] = None,
                     allversions: typing.Optional[bool] = False,
                     **kw
                     ) -> typing.Generator[Record, None, Results]:
        """
        Retrieve records matching various search criteria.

        :param community: A Zenodo community identifier.
        :param keyword: A keyword associated with relevant records.
        :param allversions: A flag signaling whether only the latest version of each record is \
        retrieved (the default) or all.
        :param kw: Additional search terms according to the Zenodo search query language.
        :return: A generator of records.
        """
        if keyword:
            kw['keywords'] = keyword
        res = Results(self, allversions=allversions, community=community, q=q(**kw))
        yield from res.next()
        while res.more:
            yield from res.next()  # pragma: no cover
        if res.error:
            return res  # pragma: no cover

    def get_github_release_info(
            self,
            repos: str,
            tag: typing.Optional[str] = None,
            bibid: typing.Optional[str] = None
    ) -> typing.Tuple[
        typing.Union[str, None],
        typing.Union[str, None],
        typing.Union[str, None],
        typing.Union[Source, None]
    ]:
        """
        Get information about a GitHub repository release that has been archived with Zenodo.

        We assume, that the DOI of the upload on Zenodo has been added to the release description
        on GitHub.

        :param repos: GitHub repository specifier in the form `org/repo`.
        :param tag: Optional version tag.
        :param bibid: Optional identifier to use for the BibTeX record.
        :return: A quadruple (tag, DOI, APA citation, BibTeX record)
        """
        def gh(args):
            return subprocess.check_output([ensure_cmd('gh')] + shlex.split(args)).decode('utf8')

        version_pattern = re.compile(r'(?P<tag>v[0-9]+\.[0-9]+(\.[0-9]+)?)')
        doi_pattern = re.compile(r'https://doi\.org/(?P<doi>10\.5281/zenodo\.[0-9]+)')

        if not tag:
            latest = gh('release list -L 1 -R {}'.format(repos))
            if not latest:
                return None, None, None, None  # pragma: no cover
            match = version_pattern.search(latest)
            assert match, latest
            tag = match.group('tag')

        info = gh('release view {} -R {}'.format(tag, repos))
        match = None
        for match in doi_pattern.finditer(info):
            pass
        if match:
            doi = match.group('doi')
            if doi:
                cit = self.records(
                    id_=ZENODO_DOI_PATTERN.match(doi).group('recid'),
                    params=dict(style='apa'),
                    headers=dict(Accept='text/x-bibliography')).strip()
                bib = self.records(
                    id_=ZENODO_DOI_PATTERN.match(doi).group('recid'),
                    params=dict(style='bibtex'),
                    headers=dict(Accept='text/x-bibliography')).strip()
                bib = Source.from_bibtex(re.sub(
                    r', (?P<field>[a-zA-Z]+)=',
                    lambda m: ',\n  {} = '.format(m.group('field')),
                    bib), _check_id=False)
                bib.id = bibid or repos.replace('/', '_')
                for field in ['abstractNote', 'month']:
                    if field in bib:
                        del bib[field]
                bib['edition'] = tag
                bib['type'] = 'Data set'
                return tag, doi, cit, bib
        return tag, None, None, None  # pragma: no cover


# Since an Api instance does some caching, we provide one instance as module global.
API = Api()


class ZenodoResolver(DatasetResolver):
    def __call__(self, loc, download_dir):
        doi = None
        m = ZENODO_DOI_PATTERN.search(loc)
        if m:
            doi = loc[m.start():m.end()]
        else:
            m = re.search(r'zenodo\.org/record(s)?/(?P<number>[0-9]+)', loc)
            if m:
                doi = ZENODO_DOI_FORMAT.format(m.group('number'))
        if doi:
            rec = API.get_record(doi=doi) or API.get_record(conceptdoi=doi)
            return rec.download(download_dir)



import zipfile

import pytest

from cldfzenodo.api import Api


def test_DatasetResolver(mocker, tmp_path, fixtures_dir):
    from pycldf.ext.discovery import get_dataset

    class Record:
        def download(self, d):
            with zipfile.ZipFile(fixtures_dir / 'petersonsouthasia-v1.1.zip', 'r') as zf:
                zf.extractall(d)
            return d

    class API:
        def get_record(self, *args, **kw):
            return Record()

    mocker.patch('cldfzenodo.api.API', API())
    assert get_dataset('https://doi.org/10.5281/zenodo.1', tmp_path)
    assert get_dataset('https://zenodo.org/record/1', tmp_path)
    assert tmp_path.joinpath('cldf-datasets-petersonsouthasia-e029fbf').exists()


def test_Api_iter_records(urlopen, mocker):
    mocker.patch('cldfzenodo.api.urllib.request.urlopen', urlopen)
    res = Api(page_size=10).iter_records(allversions=True)
    recs = []
    for i in range(15):
        recs.append(next(res))
    assert len(recs) == 15
    assert len({r.doi for r in recs}) == 10, "Should only get 10 different records"

    res = list(Api().iter_records(community='lexibank'))
    assert len(res) == 10

    res = list(Api().iter_records(keyword='x'))
    assert len(res) == 10


def test_Api_get_record(urlopen, mocker):
    mocker.patch('cldfzenodo.api.urllib.request.urlopen', urlopen)

    rec = Api().get_record(doi='10.5281/zenodo.5173799')
    assert rec.doi == '10.5281/zenodo.5173799'

    rec = Api().get_record(conceptdoi='10.5281/zenodo.5173799')
    assert rec.doi == '10.5281/zenodo.5173799'

    assert Api().get_record(conceptdoi='10.5281/zenodo.5173799', version='y') is None
    assert Api().get_record(conceptdoi='10.5281/zenodo.5173799', version='v1.1') is not None

    assert list(Api().iter_versions('10.5281/zenodo.5173799'))


def test_Api_get_github_release_info(mocker):
    class Subprocess:
        @staticmethod
        def check_output(cmd):
            if 'list' in cmd:
                return b"""TITLE                   TYPE    TAG NAME  PUBLISHED         
Glottolog Database 4.8  Latest  v4.8      about 4 months ago"""
            return """v4.8
xrotwang released this about 4 months ago

  Hammarström, Harald & Forkel, Robert & Haspelmath, Martin & Bank, Sebastian. 2023. Glottolog 4.8. Leipzig: Max      
  Planck Institute for Evolutionary Anthropology. (Available online at https://glottolog.org/)                        
                                                                                                                      
  DOI https://doi.org/10.5281/zenodo.8131084                                                                          


View on GitHub: https://github.com/glottolog/glottolog/releases/tag/v4.8""".encode('utf8')

    class Response:
        def __init__(self, what):
            self.what = what

        def read(self):
            if self.what == 'apa':
                return b'The Citation'
            return (' @misc{harald hammarström_robert forkel_martin haspelmath_sebastian bank_2023, '
                    'title={glottolog/glottolog: Glottolog database 4.8}, '
                    'DOI={10.5281/zenodo.8131084}, '
                    'abstractNote={Hammarström, Harald & Forkel, Robert & Haspelmath, Martin & Bank, Sebastian. 2023. Glottolog 4.8. Leipzig: Max Planck Institute for Evolutionary Anthropology. (Available online at https://glottolog.org)}, '
                    'publisher={Zenodo}, '
                    'author={Harald Hammarström and Robert Forkel and Martin Haspelmath and Sebastian Bank}, '
                    'year={2023}, '
                    'month={Jul} }').encode('utf8')

    def urlopen(req):
        if 'style=apa' in req.full_url:
            return Response('apa')
        return Response('bibtex')

    mocker.patch('cldfzenodo.api.subprocess', Subprocess())
    mocker.patch('cldfzenodo.api.urllib.request.urlopen', urlopen)

    tag, doi, cit, bib = Api().get_github_release_info('glottolog/glottolog', bibid='x')
    assert tag == 'v4.8'
    assert doi == '10.5281/zenodo.8131084'
    assert cit == 'The Citation'
    assert bib.id == 'x'



import pytest

from cldfzenodo.api import Api


@pytest.fixture(scope='session')
def API():  # pragma: no cover
    class CountingApi(Api):
        def __init__(self, *args, **kw):
            Api.__init__(self, *args, **kw)
            self.total_requests = 0
            self.scoped_requests = 0

        def __call__(self, *args, **kw):
            self.total_requests += 1
            self.scoped_requests += 1
            kw['_verbose'] = True
            return Api.__call__(self, *args, **kw)

        def __enter__(self):
            self.scoped_requests = 0
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    return CountingApi(page_size=5)


def get_10(iterator):  # pragma: no cover
    res = []
    for i in range(10):
        res.append(next(iterator))
    return res


@pytest.mark.integration
def test_communities(API):  # pragma: no cover
    with API:
        res = get_10(API.iter_records(community='cldf-datasets'))
    assert len(res) == 10
    assert all('cldf-datasets' in r.communities for r in res)
    assert len({r.doi for r in res}) == 10
    assert API.scoped_requests == 3


@pytest.mark.integration
def test_keyword(API):  # pragma: no cover
    with API:
        res = get_10(API.iter_records(keyword='cldf:Generic'))
    assert len(res) == 10
    assert all('cldf:Generic' in r.keywords for r in res)
    assert len({r.doi for r in res}) == 10
    assert API.scoped_requests == 2


@pytest.mark.integration
def test_record(API):  # pragma: no cover
    with API:
        rec = API.get_record(doi="10.5281/zenodo.5173799")
        cit = rec.get_citation(API)
    assert rec.id == '5173799'
    assert 'Greenhill' in cit
    assert API.scoped_requests == 2

    rec1 = API.get_record(conceptdoi="10.5281/zenodo.4691101", version="v1.1")
    assert rec1.version == 'v1.1'
    assert rec1.concept_doi == "10.5281/zenodo.4691101"

    rec2 = API.get_record(conceptdoi="10.5281/zenodo.4691101")
    assert rec2.version != 'v1.1', "There ought to be a newer version for this conceptdoi!"


@pytest.mark.integration
def test_versions(API):  # pragma: no cover
    assert len(list(API.iter_versions('10.5281/zenodo.3260727'))) > 5


@pytest.mark.integration
def test_download(API, tmp_path):  # pragma: no cover
    rec = API.get_record(doi="10.5281/zenodo.3603755")
    assert rec.download_urls, str(rec.download_urls)
    rec.download(tmp_path, unwrap=False, prefer_github=False)
    assert tmp_path.joinpath('cldf-datasets-petersonsouthasia-e029fbf').exists(), \
        [n.name for n in tmp_path.iterdir()]
    ds = rec.download_dataset(tmp_path)
    assert len(ds.objects('LanguageTable')) == 29



"""
Zenodo deposit record, as described by the DCAT metadata.
"""
import io
import re
import shutil
import typing
import pathlib
import zipfile
import tempfile
import urllib.parse
import urllib.request

import attr
import nameparser
from pycldf import iter_datasets, Source, Dataset

__all__ = ['Record', 'GithubRepos', 'ZENODO_DOI_FORMAT', 'ZENODO_DOI_PATTERN', 'get_doi']

ZENODO_DOI_PATTERN = re.compile(r"10\.5281/zenodo\.(?P<recid>[0-9]+)")
DOI_PATTERN = re.compile(r"10\.[0-9.]+/[^/]+")
ZENODO_DOI_FORMAT = '10.5281/zenodo.{}'
NS = dict(
    rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    adms="http://www.w3.org/ns/adms#",
    dc="http://purl.org/dc/elements/1.1/",
    dct="http://purl.org/dc/terms/",
    citedcat="https://w3id.org/citedcat-ap/",
    dctype="http://purl.org/dc/dcmitype/",
    dcat="http://www.w3.org/ns/dcat#",
    duv="http://www.w3.org/ns/duv#",
    foaf="http://xmlns.com/foaf/0.1/",
    frapo="http://purl.org/cerif/frapo/",
    geo="http://www.w3.org/2003/01/geo/wgs84_pos#",
    gsp="http://www.opengis.net/ont/geosparql#",
    locn="http://www.w3.org/ns/locn#",
    org="http://www.w3.org/ns/org#",
    owl="http://www.w3.org/2002/07/owl#",
    prov="http://www.w3.org/ns/prov#",
    rdfs="http://www.w3.org/2000/01/rdf-schema#",
    schema="http://schema.org/",
    skos="http://www.w3.org/2004/02/skos/core#",
    vcard="http://www.w3.org/2006/vcard/ns#",
    wdrs="http://www.w3.org/2007/05/powder-s#",
)


@attr.s
class GithubRepos:
    """
    Zenodo deposits via the GitHub-Zenodo bridge carry enough metadata to retrieve the source
    repository. This info is particularly interesting for us, because we prefer downloading
    releases from GitHub over Zenodo, due to Zenodo's rate-limiting.

    :ivar str org: GitHub organization name of the repository
    :ivar str name: repository name
    :ivar str tag: release tag of the version of the repository that was deposited on Zenodo
    """
    org = attr.ib()
    name = attr.ib()
    tag = attr.ib(default=None)

    @classmethod
    def from_url(cls, url: str) -> 'GithubRepos':
        url = urllib.parse.urlparse(url)
        if url.netloc == 'github.com':
            path = url.path.split('/')
            return cls(
                org=path[1],
                name=path[2],
                tag=path[4] if len(path) > 4 and path[3] == 'tree' else None)

    @property
    def clone_url(self) -> str:
        """
        :return: A URL suitable for passing to `git clone`.
        """
        return 'https://github.com/{0.org}/{0.name}.git'.format(self)

    @property
    def release_url(self) -> typing.Union[None, str]:
        """
        :return: The URL of a zipped release on GitHub.
        """
        if self.tag:
            return 'https://github.com/{0.org}/{0.name}/archive/refs/tags/{0.tag}.zip'.format(self)


def get_doi(doi_or_url: str) -> str:
    """

    .. code-block:: python

        >>> get_doi('10.5281/zenodo.5173799')
        '10.5281/zenodo.5173799'
        >>> get_doi("https://doi.org/10.5281/zenodo.5173799")
        '10.5281/zenodo.5173799'
        >>> get_doi("https://zenodo.org/doi/10.5281/zenodo.5173799")
        '10.5281/zenodo.5173799'
        >>> get_doi("https://zenodo.org/record/5173799")
        '10.5281/zenodo.5173799'
        >>> get_doi("https://zenodo.org/records/5173799")
        '10.5281/zenodo.5173799'
    """
    url, doi = urllib.parse.urlparse(doi_or_url), None
    if not url.netloc:
        doi = url.path
    elif url.netloc == 'zenodo.org':
        if url.path.startswith('/doi/'):
            doi = url.path.replace('/doi/', '')
        else:
            doi = ZENODO_DOI_FORMAT.format(url.path.split('/')[-1])
    elif url.netloc == 'doi.org':
        doi = url.path[1:]
    else:
        raise ValueError('Unknown DOI format')
    if not (ZENODO_DOI_PATTERN.fullmatch(doi) or DOI_PATTERN.fullmatch(doi)):
        raise ValueError('Not a DOI: "{}"'.format(doi))
    return doi


def get_creators(names):
    res = []
    for name in names:
        name = nameparser.HumanName(name)
        first = name.first
        if name.middle:
            first += ' ' + name.middle
        res.append('{}, {}'.format(name.last, first))
    return res


@attr.s
class Record:
    """
    Metadata of a Zenodo deposit

    :ivar doi: The DOI of the record.
    :ivar title: The title of the record.
    :ivar typing.List[str] creators: List of names of the creators of the record.
    """
    doi = attr.ib(
        converter=get_doi,
        validator=attr.validators.matches_re(DOI_PATTERN))
    title = attr.ib()
    creators = attr.ib(converter=get_creators, default=attr.Factory(list))
    year = attr.ib(default=None)
    license = attr.ib(default=None)
    download_urls = attr.ib(default=attr.Factory(list))
    keywords = attr.ib(default=attr.Factory(list))
    communities = attr.ib(default=attr.Factory(list), converter=lambda c: [i for i in c if i])
    github_repos = attr.ib(default=None)
    closed_access = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    version = attr.ib(default=None)
    concept_doi = attr.ib(default=None)

    def __attrs_post_init__(self):
        if not self.download_url:
            assert self.closed_access, self.doi

    @property
    def download_url(self) -> typing.Union[str, None]:
        return self.download_urls[0] if self.download_urls else None

    @property
    def id(self) -> str:
        """
        :return: The Zenodo recid.
        """
        return self.doi.replace('10.5281/zenodo.', '')

    @property
    def version_tag(self) -> typing.Union[None, str]:  # pragma: no cover
        return self.version or (self.github_repos.tag if self.github_repos else None)

    @classmethod
    def from_dict(cls, d):
        kw = dict(
            doi=d['doi'],
            title=d['metadata']['title'],
            keywords=d['metadata'].get('keywords'),
            communities=[
                dd.get('identifier', dd.get('id')) for dd in d['metadata'].get('communities', [])],
            closed_access=d['metadata']['access_right'] in {'closed', 'restricted'},
            creators=[c['name'] for c in d['metadata']['creators']],
            year=d['metadata']['publication_date'].split('-')[0],
            version=d['metadata'].get('version'),
            concept_doi=d.get('conceptdoi'),  # There are old records with "concept_rec_id" ...
            # FIXME: Check Zenodo API periodically to see whether URLs are correct now.
            download_urls=[f['links']['self'].replace('/api/', '/') for f in d.get('files')],
        )
        if 'license' in d['metadata']:
            lic = d['metadata']['license']
            kw['license'] = lic if isinstance(lic, str) else lic.get('identifier', lic.get('id'))
        for ri in d['metadata'].get('related_identifiers', []):
            if ri['relation'] == 'isSupplementTo':
                kw['github_repos'] = GithubRepos.from_url(ri['identifier'])
        return cls(**kw)

    @staticmethod
    def _download(url, dest, log=None):
        urlpath = pathlib.Path(urllib.parse.urlparse(url).path)
        with urllib.request.urlopen(url) as res:
            if res.code == 200:
                if log:
                    log.info('Downloading {}'.format(url))
                if urlpath.suffix == '.zip':
                    zipfile.ZipFile(io.BytesIO(res.read())).extractall(path=dest)
                else:
                    dest.joinpath(urlpath.name).write_bytes(res.read())

    def download(self, dest, log=None, unwrap=True, prefer_github=True) -> pathlib.Path:
        """
        Download the zipped file-content of the record to `dest`.

        :param dest:
        :param log:
        :return: The directory containing the unzipped files of the record.
        """
        dest = pathlib.Path(dest)
        is_empty = not dest.exists() or (len(list(dest.iterdir())) == 0)
        if not dest.exists():
            dest.mkdir()
        if not self.download_urls:
            raise ValueError('No downloadable resources')  # pragma: no cover
        # Preferentially download from github to not run into Zenodo's rate limit.
        if prefer_github and self.github_repos and self.github_repos.release_url:
            self._download(self.github_repos.release_url, dest, log=log)
        else:
            for url in self.download_urls:
                if url.endswith('/content'):
                    url = url[:-len('/content')]  # pragma: no cover
                self._download(url, dest, log=log)
        inner = list(dest.iterdir())
        if unwrap and is_empty and len(inner) == 1 and inner[0].is_dir():
            # Move the content of the inner-directory to dest:
            for p in inner[0].iterdir():
                shutil.move(str(p), str(dest))
            inner[0].rmdir()
        return dest

    def download_dataset(self, dest, condition=None, mdname=None, log=None) -> Dataset:
        with tempfile.TemporaryDirectory() as tmpdirname:
            for ds in iter_datasets(self.download(tmpdirname, log=log)):
                if (condition is None) or condition(ds):
                    return Dataset.from_metadata(ds.copy(dest, mdname=mdname))

    @property
    def bibtex(self) -> str:
        src = Source(
            'misc',
            self.doi.split('/')[-1].replace('.', '-'),
            author=' and '.join(self.creators),
            title=self.title,
            keywords=', '.join(self.keywords),
            publisher='Zenodo',
            year=self.year,
            doi=self.doi,
            url='https://doi.org/{}'.format(self.doi),
        )
        if self.license:
            src['copyright'] = self.license
        return src.bibtex()

    def get_citation(self, api) -> str:
        # curl -H "Accept:text/x-bibliography" "https://zenodo.org/api/records/7079637?style=apa
        return api.records(
            id_=self.id,
            params=dict(style='apa'),
            headers=dict(Accept='text/x-bibliography')).strip()

    # ---------------------------------------------------------------------------------------------
    # legacy API:
    # ---------------------------------------------------------------------------------------------
    @staticmethod
    def from_doi(doi):  # pragma: no cover
        from cldfzenodo import API
        return API.get_record(doi=doi)

    @staticmethod
    def from_concept_doi(doi, version_tag):  # pragma: no cover
        from cldfzenodo import API
        return API.get_record(conceptdoi=doi, version=version_tag)

    @property
    def citation(self):  # pragma: no cover
        from cldfzenodo import API
        return self.get_citation(API)



import warnings

from clldutils.misc import deprecated

from cldfzenodo.record import *
from cldfzenodo.api import *
from cldfzenodo import oai
from cldfzenodo import search

__version__ = '2.1.0'
# flake8: noqa

# -------------------------------------------------------------------------------------------------
# legacy API:
# -------------------------------------------------------------------------------------------------
def search_wordlists(q=None, **kw):
    deprecated('Use `API.iter_records` instead.')
    if kw:
        warnings.warn('Zenodo search API changed, custom parameters passed as kw are ignored.')
    return API.iter_records(keyword='cldf:Wordlist', allversions=True, _q=q)


def search_structuredatasets(q=None, **kw):
    deprecated('Use `API.iter_records` instead.')
    if kw:
        warnings.warn('Zenodo search API changed, custom parameters passed as kw are ignored.')
    return API.iter_records(keyword='cldf:StructureDataset', allversions=True, _q=q)


def oai_lexibank():
    deprecated('Use `API.iter_records` instead.')
    return API.iter_records(community='lexibank', allversions=True)


def oai_cldf_datasets():
    deprecated('Use `API.iter_records` instead.')
    return API.iter_records(community='cldf-datasets', allversions=True)



"""
Download a CLDF dataset from Zenodo
"""
import pathlib

from clldutils.clilib import PathType

from cldfzenodo import API


def register(parser):
    parser.add_argument(
        'doi',
        help="DOI of the dataset, starting with 10.5281, Zenodo's DOI prefix.")
    parser.add_argument(
        '--directory',
        help='Output directory (will be created if it does not exist).',
        type=PathType(type='dir', must_exist=False),
        default=pathlib.Path('.'),
    )
    parser.add_argument(
        '--version-tag',
        help="If DOI is a concept DOI (see https://help.zenodo.org/#versioning), a version tag "
             "can be specified to select a particular version of the dataset (rather than the "
             "latest one).",
        default=None,
    )
    parser.add_argument(
        '--full-deposit',
        action='store_true',
        default=False,
        help='Download all files of the deposit (rather than just the CLDF dataset).',
    )


def run(args):
    kw = dict(conceptdoi=args.doi, version=args.version_tag) \
        if args.version_tag else dict(doi=args.doi)
    rec = API.get_record(**kw)
    if not rec:
        raise ValueError('No record found')  # pragma: no cover
    rec.download(args.directory) if args.full_deposit else rec.download_dataset(args.directory)
    return 0



from cldfbench.__main__ import main


def test_download(tmp_path, mocker, recwarn):
    class Record:
        def download_dataset(self, dest):
            dest.joinpath('test').write_text('abc', encoding='utf8')

        def download(self, dest):
            dest.joinpath('test').write_text('abc', encoding='utf8')

    class Api:
        def get_record(self, *args, **kw):
            return Record()

    mocker.patch('cldfzenodo.commands.download.API', Api())
    main(['zenodo.download', 'x', '--directory', str(tmp_path)])
    assert tmp_path.joinpath('test').exists()
    main(['zenodo.download', 'x', '--full-deposit', '--directory', str(tmp_path)])
